from typing import Any

safe_builtins: Any
full_write_guard: Any

def guarded_setattr(object, name, value) -> None: ...
def guarded_delattr(object, name) -> None: ...
def safer_getattr(object, name, default: Any | None = ..., getattr=...): ...
def guarded_iter_unpack_sequence(it, spec, _getiter_) -> None: ...
def guarded_unpack_sequence(it, spec, _getiter_): ...

safe_globals: Any
